//----------------------------------------------------------------------
// Prozessor : ATmega32
// Sprache   : C
// Datum     : 14.06.2011
// Version   : 1.0
// Autor     : Cusoe, erweitert von Thilo M.
// Programm  : vereinfachte Sonnenstandsberechnung, DCF77 - Empfang und Zeitzoneneinstellung
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// Include Dateien laden
//----------------------------------------------------------------------

#include	<avr\io.h>	//AVR Register und Konstantendefinitionen
#include	<avr\interrupt.h> //AVR Interrupt Vektoren
#include	<stdio.h>
#include 	<ctype.h>
#include 	<stdint.h>
#include	<stdlib.h>
#include	<inttypes.h>
#include	<math.h>
#define SYSCLK 16000000UL
#include	<util\delay.h>
#include "defines.h"
#include "Uhr.h"
#include "lcd.c"
#include "lcd.h"	
#include "uart.c"

//----------------------------------------------------------------------
// Vorwrtsdeklarationen
//----------------------------------------------------------------------

void sun(float jd0, float dt, float la, float br);

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);	// UART-Stream festlegen
FILE lcd_str = FDEV_SETUP_STREAM(lcd_putchar, NULL, _FDEV_SETUP_WRITE); 		// LCD-Stream festlegen

//----------------------------------------------------------------------
// Konstanten
//----------------------------------------------------------------------

#define KO 			57.29577951 //Umrechnungsfaktor von Grad nach Rad und umgekehrt
#define M_2PI 		6.283185307 // Kreiszahl 2PI

//----------------------------------------------------------------------
// Statusbits (Merker)
//----------------------------------------------------------------------

#define ST_Ausgabe			0	// 1 => zyklische Ausgabe auf Display und UART freigeben
#define ST_Messung			1	// 1 => Messung starten

//----------------------------------------------------------------------
// Globale Variablen 
//----------------------------------------------------------------------

volatile uint8_t Status = 0,
				stage = 0,
				h_sec = 0,
				mt[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

float dz = 0,
      dt = 0,
	  azi = 0,
	  elev = 0;
	  
//----------------------------------------------------------------------
	  
#include "Uhr.c"
#include "ADC-Control.c"
#include "Motorcontrol.c"

//----------------------------------------------------------------------
// Zehntel-Sekunden Pause
//----------------------------------------------------------------------
void Pause_zs (unsigned char Pz)
{
	for (unsigned char i=0;i<Pz;i++)
	{
		_delay_ms(10);
	}
}

//----------------------------------------------------------------------
// Dezimaltag vom aktuellen Tag berechnen
//----------------------------------------------------------------------

float datag(int8_t st, uint8_t mi, uint8_t se)
{
	return(st / 24.0 + mi / 1440.0 + se / 86400.0);
}

//----------------------------------------------------------------------
// Anzahl Dezimaltage seit 2000 bis aktuelles Datum berechnen
//----------------------------------------------------------------------

float dezimaltage(uint8_t ta, uint8_t mo, uint8_t ja)
{
    float jahr_tage,
	      days[] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};

    jahr_tage = ta + days[mo - 1]; //Anzahl Tage im aktuellen Jahr
    if((mo < 3) && (ja % 4 == 0)) //Erst nach dem Februar hat ein Schaltjahr einen Tag mehr
    {
		jahr_tage--;
    }

    return(jahr_tage + 365 * ja + floor(ja / 4) - 0.5); //Julianische Tagzahl seit 01.01.2000 12Uhr UTC mit Bercksichtigung von Schaltjahren
}

//----------------------------------------------------------------------
// Reduzierung eines Winkel auf 0... 2PI rad, Stunden auf 0... 24 h etc.
// x=Wert der reduziert werden soll; y=Reduzierbereich z.B. auf 24h
//----------------------------------------------------------------------

float degr(float x, float y)
{
    if(x < 0)
    {
		x = x + y;
    }
    return((x - (floor(x / y) * y)));
}

//----------------------------------------------------------------------
// Vereinfachte Sonnenstandsberechnung
//----------------------------------------------------------------------

void sun(float jd0, float dt, float la, float br)
{
	float j,
	      l,
		  g,
		  a,
		  e,
		  rekt,
		  dekl,
		  t0,
		  sh,
		  t,
		  mz,
		  rz,
		  el,
		  kr;

	j = jd0 + dt;
	l = degr(4.894950420 + 0.01720279239 * j, M_2PI); //Postion der Sonne auf der Ekliptik
	g = degr(6.240040768 + 0.01720197034 * j, M_2PI); //mittlere Anomalie
	a = l + 0.03342305518 * sin(g) + 0.0003490658504 * sin(2 * g); //ekliptikale Lnge der Sonne
	e = 0.4090928042 - 0.000000006981317008 * j; //Schiefe der Ekliptik

	rekt = atan(cos(e) * sin(a) / cos(a)); //Rektaszension
	if(cos(a) < 0)
	{
		rekt = rekt + M_PI;
	}
	
	dekl = asin(sin(e) * sin(a)); //Deklination
	
	t0 = jd0 / 36525.0; //Tageszahl in julianischen Jahrhunderten
	sh = degr(6.697376 + 2400.05134 * t0 + 1.002738 * dt * 24, 24); //mittlere Sternzeit
	t = sh * 0.2617993878 + la - rekt; //Stundenwinkel der Sonne bezogen auf den Ort

	mz = cos(t) * sin(br) - tan(dekl) * cos(br); //Nenner Azimut
	rz = atan(sin(t) / mz) + M_PI; //Azimunt wird von 0 Nord gezhlt
	if(mz < 0) //wenn Nenner vom Azimut kleiner 0 dann den Winkel in den richtigen Quadranten bringen
	{
		rz = rz + M_PI;
	}
	azi = degr(rz, M_2PI); //Azimut reduzieren

	el = asin(cos(dekl) * cos(t) * cos(br) + sin(dekl) * sin(br)); //Elevation
	kr = 0.0002967059728 / tan(el + 0.179768913 / (el + 0.08918632478)); //Korrekturwert fr Refraktion
	elev = el + kr; //Elevation einschliesslich Refraktion
}

//----------------------------------------------------------------------
// Hauptprogramm
//----------------------------------------------------------------------

int main(void)
{
	stdout = stdin = &uart_str;
	stderr = &lcd_str;
	uart_init();
	init_Motorcontrol();
	ADC_init ();
	lcd_init();
	lcd_setcrsr(0x80);						// Begrungstext ausgeben
	fprintf(stderr, "  -SUNTRACKER-  ");
	_delay_ms(1);
	lcd_setcrsr(0xc0);
	fprintf(stderr, "V0.9 THM 06/2011");
	fprintf(stdout, "-=SUNTRACKER=-  V0.9 THM 06/2011\n");
	init_clock ();			// Uhr initialisieren
	_delay_ms(2000);

	SREG |= _BV(SREG_I);								// Globalen Interrupt freigeben
	
//--- zyklischer Programmablauf ------------------------------------------------------------------------
	for (;;)
	{
		//--- Display- u. UARTausgabe ------------------------------------------------------------------
		if (Status &(1<<ST_Ausgabe))
			{
				Status &= ~(1<<ST_Ausgabe);	// Displayausgabe sperren
				dz = datag(Stunde-Sommerzeit-GMT, Minute, Sekunde);
				dt = dezimaltage(Tag, Monat, (uint8_t)(Jahr-2000));

				//sun(dt, dz, laenge / KO, breite / KO);	// im Sekundentakt berechnen
				//Messung ();						// Position des Panels messen

				fprintf(stdout, "[GMT%+1d] - %02d.%02d.%04d %02d:%02d:%02d - Azimut:% 5.1f Elevation:% 5.1f X:% 4.1f Y:% 4.1f\n",(int8_t)(GMT+Sommerzeit),Tag, Monat,Jahr,Stunde,Minute,Sekunde,(azi * KO),(elev * KO),X_Messwert,Y_Messwert);
				lcd_setcrsr(0x80);
				fprintf(stderr, "Azimut:  % 6.1f%c  ",(azi * KO),0xDF);
				_delay_ms(1);
				lcd_setcrsr(0xc0);
				fprintf(stderr, "Elevation:% 5.1f%c  ",(elev * KO),0xDF);
			}
		//----------------------------------------------------------------------------------------------
		if (Status &(1<<ST_Messung))
			{
				Status &= ~(1<<ST_Messung);
				Messung ();
			}
	} 
//------------------------------------------------------------------------------------------------------	
	return 0;
}


